@@ -30,14 +30,10 @@ from typing import (
 )
 from unittest.mock import create_autospec, patch
 
-import nptyping
 import pytest
 import typing_extensions
 from sphinx.application import Sphinx
 from sphinx.config import Config
-from sphinx.testing.util import SphinxTestApp
-from sphobjinv import Inventory
-
 from sphinx_autodoc_typehints import (
     _resolve_type_guarded_imports,
     backfill_type_hints,
@@ -49,20 +45,29 @@ from sphinx_autodoc_typehints import (
     process_docstring,
 )
 
+if typing.TYPE_CHECKING:
+    from sphinx.testing.util import SphinxTestApp
+    from sphobjinv import Inventory
+
+try:
+    import nptyping
+except ImportError:
+    nptyping = None  # type: ignore[assignment]
+
 T = TypeVar("T")
 U = TypeVar("U", covariant=True)
 V = TypeVar("V", contravariant=True)
 X = TypeVar("X", str, int)
 Y = TypeVar("Y", bound=str)
 Z = TypeVar("Z", bound="A")
-S = TypeVar("S", bound="miss")  # type: ignore # miss not defined on purpose # noqa: F821
+S = TypeVar("S", bound="miss")  # type: ignore[name-defined] # miss not defined on purpose # noqa: F821
 W = NewType("W", str)
 P = typing_extensions.ParamSpec("P")
 P_args = P.args  # type:ignore[attr-defined]
 P_kwargs = P.kwargs  # type:ignore[attr-defined]
-P_co = typing_extensions.ParamSpec("P_co", covariant=True)  # type: ignore
-P_contra = typing_extensions.ParamSpec("P_contra", contravariant=True)  # type: ignore
-P_bound = typing_extensions.ParamSpec("P_bound", bound=str)  # type: ignore
+P_co = typing_extensions.ParamSpec("P_co", covariant=True)  # type: ignore[misc]
+P_contra = typing_extensions.ParamSpec("P_contra", contravariant=True)  # type: ignore[misc]
+P_bound = typing_extensions.ParamSpec("P_bound", bound=str)  # type: ignore[misc]
 
 # Mypy does not support recursive type aliases, but
 # other type checkers do.
@@ -91,7 +96,7 @@ class D(typing_extensions.Protocol):
     ...
 
 
-class E(typing_extensions.Protocol[T]):  # type: ignore #  Invariant type variable "T" used in protocol where covariant
+class E(typing_extensions.Protocol[T]):  # type: ignore[misc]
     ...
 
 
@@ -105,23 +110,23 @@ class Metaclass(type):
 
 class HintedMethods:
     @classmethod
-    def from_magic(cls: type[T]) -> T:  # type: ignore
+    def from_magic(cls: type[T]) -> T:  # type: ignore[empty-body]
         ...
 
-    def method(self: T) -> T:  # type: ignore
+    def method(self: T) -> T:  # type: ignore[empty-body]
         ...
 
 
 PY310_PLUS = sys.version_info >= (3, 10)
+PY312_PLUS = sys.version_info >= (3, 12)
 
 if sys.version_info >= (3, 9):
     AbcCallable = collections.abc.Callable  # type: ignore[type-arg]
 else:
-    # Hacks to make it work the same in old versions.
     # We could also set AbcCallable = typing.Callable and x fail the tests that
     # use AbcCallable when in versions less than 3.9.
-    class MyGenericAlias(typing._VariadicGenericAlias, _root=True):  # noqa: SC200
-        def __getitem__(self, params):
+    class MyGenericAlias(typing._VariadicGenericAlias, _root=True):  # noqa: SLF001
+        def __getitem__(self, params):  # noqa: ANN001, ANN204
             result = super().__getitem__(params)
             # Make a copy so we don't change the name of a cached annotation
             result = result.copy_with(result.__args__)
@@ -153,7 +158,7 @@ else:
         pytest.param(Callable[..., str], "typing", "Callable", (..., str), id="Callable_returntype"),
         pytest.param(Callable[[int, str], str], "typing", "Callable", (int, str, str), id="Callable_all_types"),
         pytest.param(
-            AbcCallable[[int, str], str],  # type: ignore
+            AbcCallable[[int, str], str],  # type: ignore[type-arg,misc,valid-type]
             "collections.abc",
             "Callable",
             (int, str, str),
@@ -186,176 +191,182 @@ def test_parse_annotation(annotation: Any, module: str, class_name: str, args: t
     assert (got_mod, got_cls, got_args) == (module, class_name, args)
 
 
-@pytest.mark.parametrize(
-    ("annotation", "expected_result"),
-    [
-        (str, ":py:class:`str`"),
-        (int, ":py:class:`int`"),
-        (StringIO, ":py:class:`~io.StringIO`"),
-        (FunctionType, ":py:class:`~types.FunctionType`"),
-        (ModuleType, ":py:class:`~types.ModuleType`"),
-        (type(None), ":py:obj:`None`"),
-        (type, ":py:class:`type`"),
-        (collections.abc.Callable, ":py:class:`~collections.abc.Callable`"),
-        (Type, ":py:class:`~typing.Type`"),
-        (Type[A], ":py:class:`~typing.Type`\\[:py:class:`~%s.A`]" % __name__),
-        (Any, ":py:data:`~typing.Any`"),
-        (AnyStr, ":py:data:`~typing.AnyStr`"),
-        (Generic[T], ":py:class:`~typing.Generic`\\[:py:class:`~typing.TypeVar`\\(``T``)]"),
-        (Mapping, ":py:class:`~typing.Mapping`"),
-        (
-            Mapping[T, int],  # type: ignore[valid-type]
-            ":py:class:`~typing.Mapping`\\[:py:class:`~typing.TypeVar`\\(``T``), :py:class:`int`]",
-        ),
-        (
-            Mapping[str, V],  # type: ignore[valid-type]
-            ":py:class:`~typing.Mapping`\\[:py:class:`str`, :py:class:`~typing.TypeVar`\\(``V``, contravariant=True)]",
-        ),
-        (
-            Mapping[T, U],  # type: ignore[valid-type]
-            ":py:class:`~typing.Mapping`\\[:py:class:`~typing.TypeVar`\\(``T``), "
-            ":py:class:`~typing.TypeVar`\\(``U``, covariant=True)]",
-        ),
-        (Mapping[str, bool], ":py:class:`~typing.Mapping`\\[:py:class:`str`, " ":py:class:`bool`]"),
-        (Dict, ":py:class:`~typing.Dict`"),
-        (
-            Dict[T, int],  # type: ignore[valid-type]
-            ":py:class:`~typing.Dict`\\[:py:class:`~typing.TypeVar`\\(``T``), :py:class:`int`]",
-        ),
-        (
-            Dict[str, V],  # type: ignore[valid-type]
-            ":py:class:`~typing.Dict`\\[:py:class:`str`, :py:class:`~typing.TypeVar`\\(``V``, contravariant=True)]",
-        ),
-        (
-            Dict[T, U],  # type: ignore[valid-type]
-            ":py:class:`~typing.Dict`\\[:py:class:`~typing.TypeVar`\\(``T``),"
-            " :py:class:`~typing.TypeVar`\\(``U``, covariant=True)]",
-        ),
-        (Dict[str, bool], ":py:class:`~typing.Dict`\\[:py:class:`str`, " ":py:class:`bool`]"),
-        (Tuple, ":py:data:`~typing.Tuple`"),
-        (Tuple[str, bool], ":py:data:`~typing.Tuple`\\[:py:class:`str`, " ":py:class:`bool`]"),
-        (Tuple[int, int, int], ":py:data:`~typing.Tuple`\\[:py:class:`int`, " ":py:class:`int`, :py:class:`int`]"),
-        (Tuple[str, ...], ":py:data:`~typing.Tuple`\\[:py:class:`str`, :py:data:`...<Ellipsis>`]"),
-        (Union, ":py:data:`~typing.Union`"),
-        (Union[str, bool], ":py:data:`~typing.Union`\\[:py:class:`str`, " ":py:class:`bool`]"),
-        (Union[str, bool, None], ":py:data:`~typing.Union`\\[:py:class:`str`, " ":py:class:`bool`, :py:obj:`None`]"),
-        pytest.param(
-            Union[str, Any],
-            ":py:data:`~typing.Union`\\[:py:class:`str`, " ":py:data:`~typing.Any`]",
-            marks=pytest.mark.skipif(
-                (3, 5, 0) <= sys.version_info[:3] <= (3, 5, 2), reason="Union erases the str on 3.5.0 -> 3.5.2"
-            ),
-        ),
-        (Optional[str], ":py:data:`~typing.Optional`\\[:py:class:`str`]"),
-        (Union[str, None], ":py:data:`~typing.Optional`\\[:py:class:`str`]"),
-        (
-            Optional[Union[str, bool]],
-            ":py:data:`~typing.Union`\\[:py:class:`str`, " ":py:class:`bool`, :py:obj:`None`]",
-        ),
-        (Callable, ":py:data:`~typing.Callable`"),
-        (Callable[..., int], ":py:data:`~typing.Callable`\\[:py:data:`...<Ellipsis>`, :py:class:`int`]"),
-        (Callable[[int], int], ":py:data:`~typing.Callable`\\[\\[:py:class:`int`], " ":py:class:`int`]"),
-        (
-            Callable[[int, str], bool],
-            ":py:data:`~typing.Callable`\\[\\[:py:class:`int`, " ":py:class:`str`], :py:class:`bool`]",
-        ),
-        (
-            Callable[[int, str], None],
-            ":py:data:`~typing.Callable`\\[\\[:py:class:`int`, " ":py:class:`str`], :py:obj:`None`]",
-        ),
-        (
-            Callable[[T], T],
-            ":py:data:`~typing.Callable`\\[\\[:py:class:`~typing.TypeVar`\\(``T``)],"
-            " :py:class:`~typing.TypeVar`\\(``T``)]",
-        ),
-        (
-            AbcCallable[[int, str], bool],  # type: ignore
-            ":py:class:`~collections.abc.Callable`\\[\\[:py:class:`int`, " ":py:class:`str`], :py:class:`bool`]",
-        ),
-        (Pattern, ":py:class:`~typing.Pattern`"),
-        (Pattern[str], ":py:class:`~typing.Pattern`\\[:py:class:`str`]"),
-        (IO, ":py:class:`~typing.IO`"),
-        (IO[str], ":py:class:`~typing.IO`\\[:py:class:`str`]"),
-        (Metaclass, ":py:class:`~%s.Metaclass`" % __name__),
-        (A, ":py:class:`~%s.A`" % __name__),
-        (B, ":py:class:`~%s.B`" % __name__),
-        (B[int], ":py:class:`~%s.B`\\[:py:class:`int`]" % __name__),
-        (C, ":py:class:`~%s.C`" % __name__),
-        (D, ":py:class:`~%s.D`" % __name__),
-        (E, ":py:class:`~%s.E`" % __name__),
-        (E[int], ":py:class:`~%s.E`\\[:py:class:`int`]" % __name__),
-        (W, f':py:{"class" if PY310_PLUS else "func"}:' f"`~typing.NewType`\\(``W``, :py:class:`str`)"),
-        (T, ":py:class:`~typing.TypeVar`\\(``T``)"),
-        (U, ":py:class:`~typing.TypeVar`\\(``U``, covariant=True)"),
-        (V, ":py:class:`~typing.TypeVar`\\(``V``, contravariant=True)"),
-        (X, ":py:class:`~typing.TypeVar`\\(``X``, :py:class:`str`, :py:class:`int`)"),
-        (Y, ":py:class:`~typing.TypeVar`\\(``Y``, bound= :py:class:`str`)"),
-        (Z, ":py:class:`~typing.TypeVar`\\(``Z``, bound= A)"),
-        (S, ":py:class:`~typing.TypeVar`\\(``S``, bound= miss)"),
-        # ParamSpec should behave like TypeVar, except for missing constraints
-        (P, ":py:class:`~typing.ParamSpec`\\(``P``)"),
-        (P_co, ":py:class:`~typing.ParamSpec`\\(``P_co``, covariant=True)"),
-        (P_contra, ":py:class:`~typing.ParamSpec`\\(``P_contra``, contravariant=True)"),
-        (P_bound, ":py:class:`~typing.ParamSpec`\\(``P_bound``, bound= :py:class:`str`)"),
-        # ## These test for correct internal tuple rendering, even if not all are valid Tuple types
-        # Zero-length tuple remains
-        (Tuple[()], ":py:data:`~typing.Tuple`"),
-        # Internal single tuple with simple types is flattened in the output
-        (Tuple[(int,)], ":py:data:`~typing.Tuple`\\[:py:class:`int`]"),
-        (Tuple[(int, int)], ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:class:`int`]"),
-        # Ellipsis in single tuple also gets flattened
-        (Tuple[(int, ...)], ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:data:`...<Ellipsis>`]"),
-        # Internal tuple with following additional type cannot be flattened (specific to nptyping?)
-        # These cases will fail if nptyping restructures its internal module hierarchy
-        (
-            nptyping.NDArray[nptyping.Shape["*"], nptyping.Float],
+_CASES = [
+    (str, ":py:class:`str`"),
+    (int, ":py:class:`int`"),
+    (StringIO, ":py:class:`~io.StringIO`"),
+    (FunctionType, ":py:class:`~types.FunctionType`"),
+    (ModuleType, ":py:class:`~types.ModuleType`"),
+    (type(None), ":py:obj:`None`"),
+    (type, ":py:class:`type`"),
+    (collections.abc.Callable, ":py:class:`~collections.abc.Callable`"),
+    (Type, ":py:class:`~typing.Type`"),
+    (Type[A], ":py:class:`~typing.Type`\\[:py:class:`~%s.A`]" % __name__),
+    (Any, ":py:data:`~typing.Any`"),
+    (AnyStr, ":py:data:`~typing.AnyStr`"),
+    (Generic[T], ":py:class:`~typing.Generic`\\[:py:class:`~typing.TypeVar`\\(``T``)]"),
+    (Mapping, ":py:class:`~typing.Mapping`"),
+    (
+        Mapping[T, int],  # type: ignore[valid-type]
+        ":py:class:`~typing.Mapping`\\[:py:class:`~typing.TypeVar`\\(``T``), :py:class:`int`]",
+    ),
+    (
+        Mapping[str, V],  # type: ignore[valid-type]
+        ":py:class:`~typing.Mapping`\\[:py:class:`str`, :py:class:`~typing.TypeVar`\\(``V``, contravariant=True)]",
+    ),
+    (
+        Mapping[T, U],  # type: ignore[valid-type]
+        ":py:class:`~typing.Mapping`\\[:py:class:`~typing.TypeVar`\\(``T``), "
+        ":py:class:`~typing.TypeVar`\\(``U``, covariant=True)]",
+    ),
+    (Mapping[str, bool], ":py:class:`~typing.Mapping`\\[:py:class:`str`, :py:class:`bool`]"),
+    (Dict, ":py:class:`~typing.Dict`"),
+    (
+        Dict[T, int],  # type: ignore[valid-type]
+        ":py:class:`~typing.Dict`\\[:py:class:`~typing.TypeVar`\\(``T``), :py:class:`int`]",
+    ),
+    (
+        Dict[str, V],  # type: ignore[valid-type]
+        ":py:class:`~typing.Dict`\\[:py:class:`str`, :py:class:`~typing.TypeVar`\\(``V``, contravariant=True)]",
+    ),
+    (
+        Dict[T, U],  # type: ignore[valid-type]
+        ":py:class:`~typing.Dict`\\[:py:class:`~typing.TypeVar`\\(``T``),"
+        " :py:class:`~typing.TypeVar`\\(``U``, covariant=True)]",
+    ),
+    (Dict[str, bool], ":py:class:`~typing.Dict`\\[:py:class:`str`, :py:class:`bool`]"),
+    (Tuple, ":py:data:`~typing.Tuple`"),
+    (Tuple[str, bool], ":py:data:`~typing.Tuple`\\[:py:class:`str`, :py:class:`bool`]"),
+    (Tuple[int, int, int], ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:class:`int`, :py:class:`int`]"),
+    (Tuple[str, ...], ":py:data:`~typing.Tuple`\\[:py:class:`str`, :py:data:`...<Ellipsis>`]"),
+    (Union, ":py:data:`~typing.Union`"),
+    (Union[str, bool], ":py:data:`~typing.Union`\\[:py:class:`str`, :py:class:`bool`]"),
+    (Union[str, bool, None], ":py:data:`~typing.Union`\\[:py:class:`str`, :py:class:`bool`, :py:obj:`None`]"),
+    pytest.param(Union[str, Any], ":py:data:`~typing.Union`\\[:py:class:`str`, :py:data:`~typing.Any`]"),
+    (Optional[str], ":py:data:`~typing.Optional`\\[:py:class:`str`]"),
+    (Union[str, None], ":py:data:`~typing.Optional`\\[:py:class:`str`]"),
+    (
+        Optional[Union[str, bool]],
+        ":py:data:`~typing.Union`\\[:py:class:`str`, :py:class:`bool`, :py:obj:`None`]",
+    ),
+    (Callable, ":py:data:`~typing.Callable`"),
+    (Callable[..., int], ":py:data:`~typing.Callable`\\[:py:data:`...<Ellipsis>`, :py:class:`int`]"),
+    (Callable[[int], int], ":py:data:`~typing.Callable`\\[\\[:py:class:`int`], :py:class:`int`]"),
+    (
+        Callable[[int, str], bool],
+        ":py:data:`~typing.Callable`\\[\\[:py:class:`int`, :py:class:`str`], :py:class:`bool`]",
+    ),
+    (
+        Callable[[int, str], None],
+        ":py:data:`~typing.Callable`\\[\\[:py:class:`int`, :py:class:`str`], :py:obj:`None`]",
+    ),
+    (
+        Callable[[T], T],
+        ":py:data:`~typing.Callable`\\[\\[:py:class:`~typing.TypeVar`\\(``T``)],"
+        " :py:class:`~typing.TypeVar`\\(``T``)]",
+    ),
+    (
+        AbcCallable[[int, str], bool],  # type: ignore[valid-type,misc,type-arg]
+        ":py:class:`~collections.abc.Callable`\\[\\[:py:class:`int`, :py:class:`str`], :py:class:`bool`]",
+    ),
+    (Pattern, ":py:class:`~typing.Pattern`"),
+    (Pattern[str], ":py:class:`~typing.Pattern`\\[:py:class:`str`]"),
+    (IO, ":py:class:`~typing.IO`"),
+    (IO[str], ":py:class:`~typing.IO`\\[:py:class:`str`]"),
+    (Metaclass, ":py:class:`~%s.Metaclass`" % __name__),
+    (A, ":py:class:`~%s.A`" % __name__),
+    (B, ":py:class:`~%s.B`" % __name__),
+    (B[int], ":py:class:`~%s.B`\\[:py:class:`int`]" % __name__),
+    (C, ":py:class:`~%s.C`" % __name__),
+    (D, ":py:class:`~%s.D`" % __name__),
+    (E, ":py:class:`~%s.E`" % __name__),
+    (E[int], ":py:class:`~%s.E`\\[:py:class:`int`]" % __name__),
+    (W, f":py:{'class' if PY310_PLUS else 'func'}:`~typing.NewType`\\(``W``, :py:class:`str`)"),
+    (T, ":py:class:`~typing.TypeVar`\\(``T``)"),
+    (U, ":py:class:`~typing.TypeVar`\\(``U``, covariant=True)"),
+    (V, ":py:class:`~typing.TypeVar`\\(``V``, contravariant=True)"),
+    (X, ":py:class:`~typing.TypeVar`\\(``X``, :py:class:`str`, :py:class:`int`)"),
+    (Y, ":py:class:`~typing.TypeVar`\\(``Y``, bound= :py:class:`str`)"),
+    (Z, ":py:class:`~typing.TypeVar`\\(``Z``, bound= A)"),
+    (S, ":py:class:`~typing.TypeVar`\\(``S``, bound= miss)"),
+    # ParamSpec should behave like TypeVar, except for missing constraints
+    (P, f":py:class:`~typing.ParamSpec`\\(``P``{', bound= :py:obj:`None`' if PY312_PLUS else ''})"),
+    (
+        P_co,
+        f":py:class:`~typing.ParamSpec`\\(``P_co``{', bound= :py:obj:`None`' if PY312_PLUS else ''}, covariant=True)",
+    ),
+    (
+        P_contra,
+        f":py:class:`~typing.ParamSpec`\\(``P_contra``{', bound= :py:obj:`None`' if PY312_PLUS else ''}"
+        ", contravariant=True)",
+    ),
+    (P_bound, ":py:class:`~typing.ParamSpec`\\(``P_bound``, bound= :py:class:`str`)"),
+    # ## These test for correct internal tuple rendering, even if not all are valid Tuple types
+    # Zero-length tuple remains
+    (Tuple[()], ":py:data:`~typing.Tuple`"),
+    # Internal single tuple with simple types is flattened in the output
+    (Tuple[(int,)], ":py:data:`~typing.Tuple`\\[:py:class:`int`]"),
+    (Tuple[(int, int)], ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:class:`int`]"),
+    # Ellipsis in single tuple also gets flattened
+    (Tuple[(int, ...)], ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:data:`...<Ellipsis>`]"),
+    (
+        RecList,
+        ":py:data:`~typing.Union`\\[:py:class:`int`, :py:class:`~typing.List`\\[RecList]]",
+    ),
+    (
+        MutualRecA,
+        ":py:data:`~typing.Union`\\[:py:class:`bool`, :py:class:`~typing.List`\\[MutualRecB]]",
+    ),
+]
+
+if nptyping is not None:
+    _CASES.extend(
+        [  # Internal tuple with following additional type cannot be flattened (specific to nptyping?)
+            # These cases will fail if nptyping restructures its internal module hierarchy
             (
-                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*], "
-                ":py:class:`~numpy.float64`]"
+                nptyping.NDArray[nptyping.Shape["*"], nptyping.Float],
+                (
+                    ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*], "
+                    ":py:class:`~numpy.float64`]"
+                ),
             ),
-        ),
-        (
-            nptyping.NDArray[nptyping.Shape["64"], nptyping.Float],
             (
-                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[64],"
-                " :py:class:`~numpy.float64`]"
+                nptyping.NDArray[nptyping.Shape["64"], nptyping.Float],
+                (
+                    ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[64],"
+                    " :py:class:`~numpy.float64`]"
+                ),
             ),
-        ),
-        (
-            nptyping.NDArray[nptyping.Shape["*, *"], nptyping.Float],
             (
-                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*, "
-                "*], :py:class:`~numpy.float64`]"
+                nptyping.NDArray[nptyping.Shape["*, *"], nptyping.Float],
+                (
+                    ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*, "
+                    "*], :py:class:`~numpy.float64`]"
+                ),
             ),
-        ),
-        (
-            nptyping.NDArray[nptyping.Shape["*, ..."], nptyping.Float],
-            ":py:class:`~nptyping.ndarray.NDArray`\\[:py:data:`~typing.Any`, :py:class:`~numpy.float64`]",
-        ),
-        (
-            nptyping.NDArray[nptyping.Shape["*, 3"], nptyping.Float],
             (
-                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*, 3"
-                "], :py:class:`~numpy.float64`]"
+                nptyping.NDArray[nptyping.Shape["*, ..."], nptyping.Float],
+                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:data:`~typing.Any`, :py:class:`~numpy.float64`]",
             ),
-        ),
-        (
-            nptyping.NDArray[nptyping.Shape["3, ..."], nptyping.Float],
             (
-                ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[3, "
-                "...], :py:class:`~numpy.float64`]"
+                nptyping.NDArray[nptyping.Shape["*, 3"], nptyping.Float],
+                (
+                    ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[*, 3"
+                    "], :py:class:`~numpy.float64`]"
+                ),
             ),
-        ),
-        (
-            RecList,
-            ":py:data:`~typing.Union`\\[:py:class:`int`, :py:class:`~typing.List`\\[RecList]]",
-        ),
-        (
-            MutualRecA,
-            ":py:data:`~typing.Union`\\[:py:class:`bool`, :py:class:`~typing.List`\\[MutualRecB]]",
-        ),
-    ],
-)
+            (
+                nptyping.NDArray[nptyping.Shape["3, ..."], nptyping.Float],
+                (
+                    ":py:class:`~nptyping.ndarray.NDArray`\\[:py:class:`~nptyping.base_meta_classes.Shape`\\[3, "
+                    "...], :py:class:`~numpy.float64`]"
+                ),
+            ),
+        ],
+    )
+
+
+@pytest.mark.parametrize(("annotation", "expected_result"), _CASES)
 def test_format_annotation(inv: Inventory, annotation: Any, expected_result: str) -> None:
     conf = create_autospec(Config, _annotation_globals=globals())
     result = format_annotation(annotation, conf)
@@ -376,7 +387,10 @@ def test_format_annotation(inv: Inventory, annotation: Any, expected_result: str
         if "typing" in expected_result_not_simplified:
             expected_result_not_simplified = expected_result_not_simplified.replace("~typing", "typing")
             conf = create_autospec(
-                Config, typehints_fully_qualified=True, simplify_optional_unions=False, _annotation_globals=globals()
+                Config,
+                typehints_fully_qualified=True,
+                simplify_optional_unions=False,
+                _annotation_globals=globals(),
             )
             assert format_annotation(annotation, conf) == expected_result_not_simplified
 
@@ -450,12 +464,15 @@ def set_python_path() -> None:
 @pytest.mark.sphinx("text", testroot="dummy")
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_always_document_param_types(
-    app: SphinxTestApp, status: StringIO, warning: StringIO, always_document_param_types: bool
+    app: SphinxTestApp,
+    status: StringIO,
+    warning: StringIO,
+    always_document_param_types: bool,
 ) -> None:
     set_python_path()
 
-    app.config.always_document_param_types = always_document_param_types  # type: ignore # create flag
-    app.config.autodoc_mock_imports = ["mailbox"]  # type: ignore # create flag
+    app.config.always_document_param_types = always_document_param_types  # type: ignore[attr-defined] # create flag
+    app.config.autodoc_mock_imports = ["mailbox"]  # type: ignore[attr-defined] # create flag
 
     # Prevent "document isn't included in any toctree" warnings
     for f in Path(app.srcdir).glob("*.rst"):
@@ -468,8 +485,8 @@ def test_always_document_param_types(
             .. autoclass:: dummy_module.DataClass
                 :undoc-members:
                 :special-members: __init__
-            """
-        )
+            """,
+        ),
     )
 
     app.build()
@@ -526,7 +543,7 @@ def maybe_fix_py310(expected_contents: str) -> str:
 def test_sphinx_output_future_annotations(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
 
-    app.config.master_doc = "future_annotations"  # type: ignore # create flag
+    app.config.master_doc = "future_annotations"  # type: ignore[attr-defined] # create flag
     app.build()
 
     assert "build succeeded" in status.getvalue()  # Build succeeded
@@ -567,19 +584,20 @@ def test_sphinx_output_future_annotations(app: SphinxTestApp, status: StringIO)
 @pytest.mark.sphinx("text", testroot="dummy")
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_defaults(
-    app: SphinxTestApp, status: StringIO, defaults_config_val: str, expected: str | Exception
+    app: SphinxTestApp,
+    status: StringIO,
+    defaults_config_val: str,
+    expected: str | Exception,
 ) -> None:
     set_python_path()
 
-    app.config.master_doc = "simple"  # type: ignore # create flag
-    app.config.typehints_defaults = defaults_config_val  # type: ignore # create flag
-    try:
-        app.build()
-    except Exception as e:
-        if not isinstance(expected, Exception):
-            raise
-        assert str(expected) in str(e)
+    app.config.master_doc = "simple"  # type: ignore[attr-defined] # create flag
+    app.config.typehints_defaults = defaults_config_val  # type: ignore[attr-defined] # create flag
+    if isinstance(expected, Exception):
+        with pytest.raises(Exception, match=re.escape(str(expected))):
+            app.build()
         return
+    app.build()
     assert "build succeeded" in status.getvalue()
 
     contents = (Path(app.srcdir) / "_build/text/simple.txt").read_text()
@@ -606,27 +624,27 @@ def test_sphinx_output_defaults(
     ("formatter_config_val", "expected"),
     [
         (None, ['("bool") -- foo', '("int") -- bar', '"str"']),
-        (lambda ann, conf: "Test", ["(*Test*) -- foo", "(*Test*) -- bar", "Test"]),
+        (lambda ann, conf: "Test", ["(*Test*) -- foo", "(*Test*) -- bar", "Test"]),  # noqa: ARG005
         ("some string", Exception("needs to be callable or `None`")),
     ],
 )
 @pytest.mark.sphinx("text", testroot="dummy")
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_formatter(
-    app: SphinxTestApp, status: StringIO, formatter_config_val: str, expected: tuple[str, ...] | Exception
+    app: SphinxTestApp,
+    status: StringIO,
+    formatter_config_val: str,
+    expected: tuple[str, ...] | Exception,
 ) -> None:
     set_python_path()
 
-    app.config.master_doc = "simple"  # type: ignore # create flag
-    app.config.typehints_formatter = formatter_config_val  # type: ignore # create flag
-    try:
-        app.build()
-    except Exception as e:
-        if not isinstance(expected, Exception):
-            raise
-        assert str(expected) in str(e)
+    app.config.master_doc = "simple"  # type: ignore[attr-defined] # create flag
+    app.config.typehints_formatter = formatter_config_val  # type: ignore[attr-defined] # create flag
+    if isinstance(expected, Exception):
+        with pytest.raises(Exception, match=re.escape(str(expected))):
+            app.build()
         return
-    assert not isinstance(expected, Exception), "Expected app.build() to raise exception, but it didn’t"
+    app.build()
     assert "build succeeded" in status.getvalue()
 
     contents = (Path(app.srcdir) / "_build/text/simple.txt").read_text()
@@ -735,8 +753,8 @@ def test_bound_class_method(method: FunctionType) -> None:
 def test_syntax_error_backfill() -> None:
     # Regression test for #188
     # fmt: off
-    func = (  # Note: line break here is what previously led to SyntaxError in process_docstring
-        lambda x: x)
+    def func(x):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN202
+        return x
     # fmt: on
     backfill_type_hints(func, "func")
 
@@ -744,7 +762,7 @@ def test_syntax_error_backfill() -> None:
 @pytest.mark.sphinx("text", testroot="resolve-typing-guard")
 def test_resolve_typing_guard_imports(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
     set_python_path()
-    app.config.autodoc_mock_imports = ["viktor"]  # type: ignore # create flag
+    app.config.autodoc_mock_imports = ["viktor"]  # type: ignore[attr-defined] # create flag
     app.build()
     assert "build succeeded" in status.getvalue()
     err = warning.getvalue()
@@ -772,12 +790,12 @@ def test_no_source_code_type_guard() -> None:
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_formatter_no_use_rtype(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
-    app.config.master_doc = "simple_no_use_rtype"  # type: ignore # create flag
-    app.config.typehints_use_rtype = False  # type: ignore
+    app.config.master_doc = "simple_no_use_rtype"  # type: ignore[attr-defined]  # create flag
+    app.config.typehints_use_rtype = False  # type: ignore[attr-defined]
     app.build()
     assert "build succeeded" in status.getvalue()
     text_path = Path(app.srcdir) / "_build" / "text" / "simple_no_use_rtype.txt"
-    text_contents = text_path.read_text().replace("–", "--")
+    text_contents = text_path.read_text().replace("–", "--")  # noqa: RUF001 # keep ambiguous EN DASH
     expected_contents = """\
     Simple Module
     *************
@@ -837,12 +855,12 @@ def test_sphinx_output_formatter_no_use_rtype(app: SphinxTestApp, status: String
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_with_use_signature(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
-    app.config.master_doc = "simple"  # type: ignore # create flag
-    app.config.typehints_use_signature = True  # type: ignore
+    app.config.master_doc = "simple"  # type: ignore[attr-defined] # create flag
+    app.config.typehints_use_signature = True  # type: ignore[attr-defined]
     app.build()
     assert "build succeeded" in status.getvalue()
     text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
-    text_contents = text_path.read_text().replace("–", "--")
+    text_contents = text_path.read_text().replace("–", "--")  # noqa: RUF001 # keep ambiguous EN DASH
     expected_contents = """\
     Simple Module
     *************
@@ -866,12 +884,12 @@ def test_sphinx_output_with_use_signature(app: SphinxTestApp, status: StringIO)
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_with_use_signature_return(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
-    app.config.master_doc = "simple"  # type: ignore # create flag
-    app.config.typehints_use_signature_return = True  # type: ignore
+    app.config.master_doc = "simple"  # type: ignore[attr-defined] # create flag
+    app.config.typehints_use_signature_return = True  # type: ignore[attr-defined]
     app.build()
     assert "build succeeded" in status.getvalue()
     text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
-    text_contents = text_path.read_text().replace("–", "--")
+    text_contents = text_path.read_text().replace("–", "--")  # noqa: RUF001 # keep ambiguous EN DASH
     expected_contents = """\
     Simple Module
     *************
@@ -895,13 +913,13 @@ def test_sphinx_output_with_use_signature_return(app: SphinxTestApp, status: Str
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_sphinx_output_with_use_signature_and_return(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
-    app.config.master_doc = "simple"  # type: ignore # create flag
-    app.config.typehints_use_signature = True  # type: ignore
-    app.config.typehints_use_signature_return = True  # type: ignore
+    app.config.master_doc = "simple"  # type: ignore[attr-defined] # create flag
+    app.config.typehints_use_signature = True  # type: ignore[attr-defined]
+    app.config.typehints_use_signature_return = True  # type: ignore[attr-defined]
     app.build()
     assert "build succeeded" in status.getvalue()
     text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
-    text_contents = text_path.read_text().replace("–", "--")
+    text_contents = text_path.read_text().replace("–", "--")  # noqa: RUF001 # keep ambiguous EN DASH
     expected_contents = """\
     Simple Module
     *************
@@ -925,12 +943,12 @@ def test_sphinx_output_with_use_signature_and_return(app: SphinxTestApp, status:
 @patch("sphinx.writers.text.MAXWIDTH", 2000)
 def test_default_annotation_without_typehints(app: SphinxTestApp, status: StringIO) -> None:
     set_python_path()
-    app.config.master_doc = "without_complete_typehints"  # type: ignore # create flag
-    app.config.typehints_defaults = "comma"  # type: ignore
+    app.config.master_doc = "without_complete_typehints"  # type: ignore[attr-defined]# create flag
+    app.config.typehints_defaults = "comma"  # type: ignore[attr-defined]
     app.build()
     assert "build succeeded" in status.getvalue()
     text_path = Path(app.srcdir) / "_build" / "text" / "without_complete_typehints.txt"
-    text_contents = text_path.read_text().replace("–", "--")
+    text_contents = text_path.read_text().replace("–", "--")  # noqa: RUF001 # keep ambiguous EN DASH
     expected_contents = """\
     Simple Module
     *************
